@@ -15,13 +15,16 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  * 02110-1301, USA.
+ *
+ * You can also choose to distribute this program under the terms of
+ * the Unmodified Binary Distribution Licence (as given in the file
+ * COPYING.UBDL), provided that you have satisfied its requirements.
  */
 
-FILE_LICENCE ( GPL2_OR_LATER );
+FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 
 #include <errno.h>
 #include <ipxe/bofm.h>
-#include <ipxe/init.h>
 #include <ipxe/efi/efi.h>
 #include <ipxe/efi/efi_pci.h>
 #include <ipxe/efi/efi_driver.h>
@@ -156,49 +159,37 @@ static EFI_GUID bofm2_protocol_guid =
 /**
  * Check if device is supported
  *
- * @v driver		EFI driver
- * @v device		EFI device
- * @v child		Path to child device, if any
- * @ret efirc		EFI status code
+ * @v device		EFI device handle
+ * @ret rc		Return status code
  */
-static EFI_STATUS EFIAPI
-efi_bofm_supported ( EFI_DRIVER_BINDING_PROTOCOL *driver,
-		     EFI_HANDLE device,
-		     EFI_DEVICE_PATH_PROTOCOL *child ) {
-	struct efi_driver *efidrv =
-		container_of ( driver, struct efi_driver, driver );
+static int efi_bofm_supported ( EFI_HANDLE device ) {
 	EFI_BOOT_SERVICES *bs = efi_systab->BootServices;
+	struct pci_device pci;
 	union {
 		IBM_BOFM_DRIVER_CONFIGURATION_PROTOCOL *bofm1;
 		void *interface;
 	} bofm1;
-	struct efi_pci_device *efipci;
 	EFI_STATUS efirc;
 	int rc;
 
-	DBGCP ( efidrv, "EFIBOFM DRIVER_SUPPORTED %p (%p)\n", device, child );
-
-	/* Create corresponding PCI device, if any */
-	efipci = efipci_create ( efidrv, device );
-	if ( ! efipci ) {
-		rc = -ENOTSUP;
-		goto err_not_pci;
-	}
+	/* Get PCI device information */
+	if ( ( rc = efipci_info ( device, &pci ) ) != 0 )
+		return rc;
 
 	/* Look for a BOFM driver */
-	if ( ( rc = bofm_find_driver ( &efipci->pci ) ) != 0 ) {
-		DBGCP ( efidrv, "EFIBOFM " PCI_FMT " has no driver\n",
-			PCI_ARGS ( &efipci->pci ) );
-		goto err_no_driver;
+	if ( ( rc = bofm_find_driver ( &pci ) ) != 0 ) {
+		DBGCP ( device, "EFIBOFM %p %s has no driver\n",
+			device, efi_handle_name ( device ) );
+		return rc;
 	}
 
 	/* Locate BOFM protocol */
 	if ( ( efirc = bs->LocateProtocol ( &bofm1_protocol_guid, NULL,
 					    &bofm1.interface ) ) != 0 ) {
 		rc = -EEFI ( efirc );
-		DBGC ( efidrv, "EFIBOFM " PCI_FMT " cannot find BOFM "
-		       "protocol\n", PCI_ARGS ( &efipci->pci ) );
-		goto err_not_bofm;
+		DBGC ( device, "EFIBOFM %p %s cannot find BOFM protocol\n",
+		       device, efi_handle_name ( device ) );
+		return rc;
 	}
 
 	/* Register support for this device */
@@ -207,42 +198,25 @@ efi_bofm_supported ( EFI_DRIVER_BINDING_PROTOCOL *driver,
 						      0x00 /* No iSCSI */,
 						      0x02 /* Version */ ))!=0){
 		rc = -EEFI ( efirc );
-		DBGC ( efidrv, "EFIBOFM " PCI_FMT " could not register "
-		       "support: %s\n", PCI_ARGS ( &efipci->pci ),
-		       strerror ( rc ) );
-		goto err_cannot_register;
+		DBGC ( device, "EFIBOFM %p %s could not register support: %s\n",
+		       device, efi_handle_name ( device ), strerror ( rc ) );
+		return rc;
 	}
 
-	DBGC ( efidrv, "EFIBOFM " PCI_FMT " is supported by driver \"%s\"\n",
-	       PCI_ARGS ( &efipci->pci ), efipci->pci.id->name );
-
-	/* Destroy temporary PCI device */
-	efipci_destroy ( efidrv, efipci );
-
+	DBGC ( device, "EFIBOFM %p %s has driver \"%s\"\n",
+	       device, efi_handle_name ( device ), pci.id->name );
 	return 0;
-
- err_cannot_register:
- err_not_bofm:
- err_no_driver:
-	efipci_destroy ( efidrv, efipci );
- err_not_pci:
-	return EFIRC ( rc );
 }
 
 /**
  * Attach driver to device
  *
- * @v driver		EFI driver
- * @v device		EFI device
- * @v child		Path to child device, if any
- * @ret efirc		EFI status code
+ * @v efidev		EFI device
+ * @ret rc		Return status code
  */
-static EFI_STATUS EFIAPI efi_bofm_start ( EFI_DRIVER_BINDING_PROTOCOL *driver,
-				   EFI_HANDLE device,
-				   EFI_DEVICE_PATH_PROTOCOL *child ) {
-	struct efi_driver *efidrv =
-		container_of ( driver, struct efi_driver, driver );
+static int efi_bofm_start ( struct efi_device *efidev ) {
 	EFI_BOOT_SERVICES *bs = efi_systab->BootServices;
+	EFI_HANDLE device = efidev->device;
 	union {
 		IBM_BOFM_DRIVER_CONFIGURATION_PROTOCOL *bofm1;
 		void *interface;
@@ -251,75 +225,66 @@ static EFI_STATUS EFIAPI efi_bofm_start ( EFI_DRIVER_BINDING_PROTOCOL *driver,
 		IBM_BOFM_DRIVER_CONFIGURATION_PROTOCOL2 *bofm2;
 		void *interface;
 	} bofm2;
-	struct efi_pci_device *efipci;
+	struct pci_device pci;
 	IBM_BOFM_TABLE *bofmtab;
 	IBM_BOFM_TABLE *bofmtab2;
 	int bofmrc;
 	EFI_STATUS efirc;
 	int rc;
 
-	DBGCP ( efidrv, "EFIBOFM DRIVER_START %p (%p)\n", device, child );
-
-	/* Create corresponding PCI device */
-	efipci = efipci_create ( efidrv, device );
-	if ( ! efipci ) {
-		rc = -ENOMEM;
-		goto err_create;
-	}
-
-	/* Enable PCI device */
-	if ( ( rc = efipci_enable ( efipci ) ) != 0 )
-		goto err_enable;
+	/* Open PCI device, if possible */
+	if ( ( rc = efipci_open ( device, EFI_OPEN_PROTOCOL_GET_PROTOCOL,
+				  &pci ) ) != 0 )
+		goto err_open;
 
 	/* Locate BOFM protocol */
 	if ( ( efirc = bs->LocateProtocol ( &bofm1_protocol_guid, NULL,
 					    &bofm1.interface ) ) != 0 ) {
 		rc = -EEFI ( efirc );
-		DBGC ( efidrv, "EFIBOFM " PCI_FMT " cannot find BOFM "
-		       "protocol\n", PCI_ARGS ( &efipci->pci ) );
+		DBGC ( device, "EFIBOFM %p %s cannot find BOFM protocol\n",
+		       device, efi_handle_name ( device ) );
 		goto err_locate_bofm;
 	}
 	bofmtab = &bofm1.bofm1->BofmTable;
-	DBGC ( efidrv, "EFIBOFM " PCI_FMT " found version 1 BOFM table at "
-	       "%p+%04x\n", PCI_ARGS ( &efipci->pci ), bofmtab,
+	DBGC ( device, "EFIBOFM %p %s found version 1 BOFM table at %p+%04x\n",
+	       device, efi_handle_name ( device ), bofmtab,
 	       bofmtab->Parameters.Length );
 
 	/* Locate BOFM2 protocol, if available */
 	if ( ( efirc = bs->LocateProtocol ( &bofm2_protocol_guid, NULL,
 					    &bofm2.interface ) ) == 0 ) {
 		bofmtab2 = &bofm2.bofm2->BofmTable;
-		DBGC ( efidrv, "EFIBOFM " PCI_FMT " found version 2 BOFM table "
-		       "at %p+%04x\n", PCI_ARGS ( &efipci->pci ), bofmtab2,
-		       bofmtab2->Parameters.Length );
+		DBGC ( device, "EFIBOFM %p %s found version 2 BOFM table at "
+		       "%p+%04x\n", device, efi_handle_name ( device ),
+		       bofmtab2, bofmtab2->Parameters.Length );
 		assert ( bofm2.bofm2->RegisterSupport ==
 			 bofm1.bofm1->RegisterSupport );
 	} else {
-		DBGC ( efidrv, "EFIBOFM " PCI_FMT " cannot find BOFM2 "
-		       "protocol\n", PCI_ARGS ( &efipci->pci ) );
+		DBGC ( device, "EFIBOFM %p %s cannot find BOFM2 protocol\n",
+		       device, efi_handle_name ( device ) );
 		/* Not a fatal error; may be a BOFM1-only system */
 		bofmtab2 = NULL;
 	}
 
 	/* Process BOFM table */
-	DBGC2 ( efidrv, "EFIBOFM " PCI_FMT " version 1 before processing:\n",
-		PCI_ARGS ( &efipci->pci ) );
-	DBGC2_HD ( efidrv, bofmtab, bofmtab->Parameters.Length );
+	DBGC2 ( device, "EFIBOFM %p %s version 1 before processing:\n",
+		device, efi_handle_name ( device ) );
+	DBGC2_HD ( device, bofmtab, bofmtab->Parameters.Length );
 	if ( bofmtab2 ) {
-		DBGC2 ( efidrv, "EFIBOFM " PCI_FMT " version 2 before "
-			"processing:\n", PCI_ARGS ( &efipci->pci ) );
-		DBGC2_HD ( efidrv, bofmtab2, bofmtab2->Parameters.Length );
+		DBGC2 ( device, "EFIBOFM %p %s version 2 before processing:\n",
+			device, efi_handle_name ( device ) );
+		DBGC2_HD ( device, bofmtab2, bofmtab2->Parameters.Length );
 	}
-	bofmrc = bofm ( virt_to_user ( bofmtab2 ? bofmtab2 : bofmtab ),
-			&efipci->pci );
-	DBGC ( efidrv, "EFIBOFM " PCI_FMT " status %08x\n",
-	       PCI_ARGS ( &efipci->pci ), bofmrc );
-	DBGC2 ( efidrv, "EFIBOFM " PCI_FMT " version 1 after processing:\n",
-		PCI_ARGS ( &efipci->pci ) );
-	DBGC2_HD ( efidrv, bofmtab, bofmtab->Parameters.Length );
+	bofmrc = bofm ( virt_to_user ( bofmtab2 ? bofmtab2 : bofmtab ), &pci );
+	DBGC ( device, "EFIBOFM %p %s status %08x\n",
+	       device, efi_handle_name ( device ), bofmrc );
+	DBGC2 ( device, "EFIBOFM %p %s version 1 after processing:\n",
+		device, efi_handle_name ( device ) );
+	DBGC2_HD ( device, bofmtab, bofmtab->Parameters.Length );
 	if ( bofmtab2 ) {
-		DBGC2 ( efidrv, "EFIBOFM " PCI_FMT " version 2 after "
-			"processing:\n", PCI_ARGS ( &efipci->pci ) );
-		DBGC2_HD ( efidrv, bofmtab2, bofmtab2->Parameters.Length );
+		DBGC2 ( device, "EFIBOFM %p %s version 2 after processing:\n",
+			device, efi_handle_name ( device ) );
+		DBGC2_HD ( device, bofmtab2, bofmtab2->Parameters.Length );
 	}
 
 	/* Return BOFM status */
@@ -327,93 +292,47 @@ static EFI_STATUS EFIAPI efi_bofm_start ( EFI_DRIVER_BINDING_PROTOCOL *driver,
 		if ( ( efirc = bofm2.bofm2->SetStatus ( bofm2.bofm2, device,
 							FALSE, bofmrc ) ) != 0){
 			rc = -EEFI ( efirc );
-			DBGC ( efidrv, "EFIBOFM " PCI_FMT " could not set "
-			       "BOFM2 status: %s\n", PCI_ARGS ( &efipci->pci ),
-			       strerror ( rc ) );
+			DBGC ( device, "EFIBOFM %p %s could not set BOFM2 "
+			       "status: %s\n", device,
+			       efi_handle_name ( device ), strerror ( rc ) );
 			goto err_set_status;
 		}
 	} else {
 		if ( ( efirc = bofm1.bofm1->SetStatus ( bofm1.bofm1, device,
 							FALSE, bofmrc ) ) != 0){
 			rc = -EEFI ( efirc );
-			DBGC ( efidrv, "EFIBOFM " PCI_FMT " could not set "
-			       "BOFM status: %s\n", PCI_ARGS ( &efipci->pci ),
-			       strerror ( rc ) );
+			DBGC ( device, "EFIBOFM %p %s could not set BOFM "
+			       "status: %s\n", device,
+			       efi_handle_name ( device ), strerror ( rc ) );
 			goto err_set_status;
 		}
 	}
 
-	/* Destroy the PCI device anyway; we have no further use for it */
-	efipci_destroy ( efidrv, efipci );
-
 	/* BOFM (ab)uses the "start" method to mean "process and exit" */
-	return EFI_NOT_READY;
+	rc = -EAGAIN;
 
  err_set_status:
  err_locate_bofm:
- err_enable:
-	efipci_destroy ( efidrv, efipci );
- err_create:
-	return EFIRC ( rc );
+	efipci_close ( device );
+ err_open:
+	return rc;
 }
 
 /**
  * Detach driver from device
  *
- * @v driver		EFI driver
  * @v device		EFI device
- * @v num_children	Number of child devices
- * @v children		List of child devices
- * @ret efirc		EFI status code
  */
-static EFI_STATUS EFIAPI efi_bofm_stop ( EFI_DRIVER_BINDING_PROTOCOL *driver,
-					 EFI_HANDLE device, UINTN num_children,
-					 EFI_HANDLE *children ) {
-	struct efi_driver *efidrv =
-		container_of ( driver, struct efi_driver, driver );
+static void efi_bofm_stop ( struct efi_device *efidev __unused ) {
 
-	DBGCP ( efidrv, "EFIBOFM DRIVER_STOP %p (%ld %p)\n",
-		device, ( ( unsigned long ) num_children ), children );
-
-	return 0;
+	/* Should never happen */
+	assert ( 0 );
 }
 
 /** EFI BOFM driver */
-static struct efi_driver efi_bofm_driver =
-	EFI_DRIVER_INIT ( "BOFM",
-			  efi_bofm_supported, efi_bofm_start, efi_bofm_stop );
-
-/**
- * Install EFI BOFM driver
- *
- */
-static void efi_bofm_driver_startup ( void ) {
-	struct efi_driver *efidrv = &efi_bofm_driver;
-	int rc;
-
-	/* Install driver */
-	if ( ( rc = efi_driver_install ( efidrv ) ) != 0 ) {
-		DBGC ( efidrv, "EFIBOFM could not install driver: %s\n",
-		       strerror ( rc ) );
-		return;
-	}
-
-	DBGC ( efidrv, "EFIBOFM driver installed\n" );
-}
-
-/**
- * Shut down EFI BOFM driver
- *
- * @v booting		System is shutting down for OS boot
- */
-static void efi_bofm_driver_shutdown ( int booting __unused ) {
-	struct efi_driver *efidrv = &efi_bofm_driver;
-
-	efi_driver_uninstall ( efidrv );
-}
-
-/** EFI BOFM startup function */
-struct startup_fn startup_bofm __startup_fn ( STARTUP_EARLY ) = {
-	.startup = efi_bofm_driver_startup,
-	.shutdown = efi_bofm_driver_shutdown,
+struct efi_driver efi_bofm_driver __efi_driver ( EFI_DRIVER_EARLY ) = {
+	.name = "BOFM",
+	.supported = efi_bofm_supported,
+	.start = efi_bofm_start,
+	.stop = efi_bofm_stop,
 };
